class CallbackInfo {
    constructor(parent, thiz, data, args, fn) {
        this.parent = parent;
        this.thiz = thiz;
        this.data = data;
        this.args = args;
        this.fn = fn;
    }
    getNewTarget(envObject) {
        const thiz = this.thiz;
        if (thiz == null || thiz.constructor == null)
            return 0;
        return thiz instanceof this.fn ? envObject.ensureHandleId(thiz.constructor) : 0;
    }
}
class CallbackInfoStack {
    constructor() {
        this.current = null;
    }
    pop() {
        const current = this.current;
        if (current === null)
            return;
        this.current = current.parent;
    }
    push(thiz, data, args, fn) {
        const info = new CallbackInfo(this.current, thiz, data, args, fn);
        this.current = info;
        return info;
    }
    dispose() {
        this.current = null;
    }
}

class Store {
    constructor(capacity) {
        this._values = [undefined];
        this._values.length = capacity;
        this._size = 1;
        this._freeList = [];
    }
    add(value) {
        let id;
        if (this._freeList.length) {
            id = this._freeList.shift();
        }
        else {
            id = this._size;
            this._size++;
            const capacity = this._values.length;
            if (id >= capacity) {
                this._values.length = Math.ceil(capacity * 1.5);
            }
        }
        value.id = id;
        this._values[id] = value;
    }
    get(id) {
        return this._values[id];
    }
    has(id) {
        return this._values[id] !== undefined;
    }
    remove(id) {
        const value = this._values[id];
        if (value) {
            value.id = 0;
            this._values[id] = undefined;
            this._freeList.push(Number(id));
        }
    }
    dispose() {
        for (let i = 1; i < this._size; ++i) {
            const value = this._values[i];
            value === null || value === void 0 ? void 0 : value.dispose();
        }
        this._values = [undefined];
        this._size = 1;
        this._freeList = [];
    }
}

class EnvStore extends Store {
    constructor() {
        super(2);
    }
}

const supportNewFunction = (function () {
    let f;
    try {
        f = new Function();
    }
    catch (_) {
        return false;
    }
    return typeof f === 'function';
})();
const _global = (function () {
    if (typeof globalThis !== 'undefined')
        return globalThis;
    let g = (function () { return this; })();
    if (!g && supportNewFunction) {
        try {
            g = new Function('return this')();
        }
        catch (_) { }
    }
    if (!g) {
        if (typeof __webpack_public_path__ === 'undefined') {
            if (typeof global !== 'undefined')
                return global;
        }
        if (typeof window !== 'undefined')
            return window;
        if (typeof self !== 'undefined')
            return self;
    }
    return g;
})();
const emptyException = new Error();
class TryCatch {
    constructor() {
        this._exception = emptyException;
    }
    hasCaught() {
        return this._exception !== emptyException;
    }
    exception() {
        return this._exception;
    }
    setError(err) {
        this._exception = err;
    }
    reset() {
        this._exception = emptyException;
    }
    extractException() {
        const e = this._exception;
        this.reset();
        return e;
    }
}
const canSetFunctionName = (function () {
    var _a;
    try {
        return Boolean((_a = Object.getOwnPropertyDescriptor(Function.prototype, 'name')) === null || _a === void 0 ? void 0 : _a.configurable);
    }
    catch (_) {
        return false;
    }
})();
const supportReflect = typeof Reflect === 'object';
const supportFinalizer = (typeof FinalizationRegistry !== 'undefined') && (typeof WeakRef !== 'undefined');
const supportBigInt = typeof BigInt !== 'undefined';
function isReferenceType(v) {
    return (typeof v === 'object' && v !== null) || typeof v === 'function';
}
const _setImmediate = typeof setImmediate === 'function'
    ? setImmediate
    : function (f) {
        if (typeof f !== 'function')
            return;
        let channel = new MessageChannel();
        channel.port1.onmessage = function () {
            channel.port1.onmessage = null;
            channel = undefined;
            f();
        };
        channel.port2.postMessage(null);
    };
const _require = (function () {
    let nativeRequire;
    if (typeof __webpack_public_path__ !== 'undefined') {
        nativeRequire = (function () {
            return typeof __non_webpack_require__ !== 'undefined' ? __non_webpack_require__ : undefined;
        })();
    }
    else {
        nativeRequire = (function () {
            return typeof __webpack_public_path__ !== 'undefined' ? (typeof __non_webpack_require__ !== 'undefined' ? __non_webpack_require__ : undefined) : (typeof require !== 'undefined' ? require : undefined);
        })();
    }
    return nativeRequire;
})();
const Buffer = (function () {
    if (typeof _global.Buffer === 'function')
        return _global.Buffer;
    try {
        return _require('buffer').Buffer;
    }
    catch (_) { }
    return undefined;
})();

class Handle {
    constructor(id, value) {
        this.id = id;
        this.value = value;
    }
    init(id, value) {
        this.id = id;
        this.value = value;
    }
    data(envObject) {
        return envObject.getObjectBinding(this.value).data;
    }
    isEmpty() {
        return this.id === 0;
    }
    isNumber() {
        return !this.isEmpty() && typeof this.value === 'number';
    }
    isBigInt() {
        return !this.isEmpty() && typeof this.value === 'bigint';
    }
    isString() {
        return !this.isEmpty() && typeof this.value === 'string';
    }
    isFunction() {
        return !this.isEmpty() && typeof this.value === 'function';
    }
    isExternal() {
        return !this.isEmpty() && (isReferenceType(this.value) && Object.getPrototypeOf(this.value) === null);
    }
    isObject() {
        return !this.isEmpty() && typeof this.value === 'object' && this.value !== null;
    }
    isArray() {
        return !this.isEmpty() && Array.isArray(this.value);
    }
    isArrayBuffer() {
        return !this.isEmpty() && (this.value instanceof ArrayBuffer);
    }
    isTypedArray() {
        return !this.isEmpty() && (ArrayBuffer.isView(this.value)) && !(this.value instanceof DataView);
    }
    isBuffer() {
        return !this.isEmpty() && typeof Buffer === 'function' && Buffer.isBuffer(this.value);
    }
    isDataView() {
        return !this.isEmpty() && (this.value instanceof DataView);
    }
    isDate() {
        return !this.isEmpty() && (this.value instanceof Date);
    }
    isPromise() {
        return !this.isEmpty() && (this.value instanceof Promise);
    }
    isBoolean() {
        return !this.isEmpty() && typeof this.value === 'boolean';
    }
    isUndefined() {
        return !this.isEmpty() && this.value === undefined;
    }
    isSymbol() {
        return !this.isEmpty() && typeof this.value === 'symbol';
    }
    isNull() {
        return !this.isEmpty() && this.value === null;
    }
    dispose() {
        if (this.id === 0)
            return;
        this.id = 0;
        this.value = undefined;
    }
}
class ConstHandle extends Handle {
    constructor(id, value) {
        super(id, value);
    }
    dispose() { }
}
function External() {
    Object.setPrototypeOf(this, null);
}
External.prototype = null;
class HandleStore {
    constructor() {
        this._values = [
            undefined,
            HandleStore.UNDEFINED,
            HandleStore.NULL,
            HandleStore.FALSE,
            HandleStore.TRUE,
            HandleStore.GLOBAL
        ];
        this._next = HandleStore.MIN_ID;
    }
    push(value) {
        let h;
        const next = this._next;
        const values = this._values;
        if (next < values.length) {
            h = values[next];
            h.init(next, value);
        }
        else {
            h = new Handle(next, value);
            values[next] = h;
        }
        this._next++;
        return h;
    }
    erase(start, end) {
        this._next = start;
        const values = this._values;
        for (let i = start; i < end; ++i) {
            values[i].dispose();
        }
    }
    get(id) {
        return this._values[id];
    }
    swap(a, b) {
        const values = this._values;
        const h = values[a];
        values[a] = values[b];
        values[a].id = Number(a);
        values[b] = h;
        h.id = Number(b);
    }
    dispose() {
        this._values.length = HandleStore.MIN_ID;
        this._next = HandleStore.MIN_ID;
    }
}
HandleStore.UNDEFINED = new ConstHandle(1, undefined);
HandleStore.NULL = new ConstHandle(2, null);
HandleStore.FALSE = new ConstHandle(3, false);
HandleStore.TRUE = new ConstHandle(4, true);
HandleStore.GLOBAL = new ConstHandle(5, _global);
HandleStore.MIN_ID = 6;

class HandleScope {
    constructor(handleStore, id, parentScope, start, end = start) {
        this.handleStore = handleStore;
        this.id = id;
        this.parent = parentScope;
        this.child = null;
        if (parentScope !== null)
            parentScope.child = this;
        this.start = start;
        this.end = end;
        this._escapeCalled = false;
    }
    add(value) {
        const h = this.handleStore.push(value);
        this.end++;
        return h;
    }
    addExternal(envObject, data) {
        const value = new External();
        const h = envObject.ctx.handleStore.push(value);
        const binding = envObject.initObjectBinding(value);
        binding.data = data;
        this.end++;
        return h;
    }
    dispose() {
        if (this.start === this.end)
            return;
        this.handleStore.erase(this.start, this.end);
    }
    escape(handle) {
        if (this._escapeCalled)
            return null;
        this._escapeCalled = true;
        if (handle < this.start || handle >= this.end) {
            return null;
        }
        this.handleStore.swap(handle, this.start);
        const h = this.handleStore.get(this.start);
        this.start++;
        this.parent.end++;
        return h;
    }
    escapeCalled() {
        return this._escapeCalled;
    }
}

class ScopeStore {
    constructor() {
        this._rootScope = new HandleScope(null, 0, null, 1, HandleStore.MIN_ID);
        this.currentScope = this._rootScope;
    }
    get(id) {
        id = Number(id);
        let scope = this.currentScope;
        while (scope !== this._rootScope) {
            if (scope.id === id) {
                return scope;
            }
            scope = scope.parent;
        }
        return undefined;
    }
    openScope(envObject) {
        const currentScope = this.currentScope;
        let scope = currentScope.child;
        if (scope !== null) {
            scope.start = scope.end = currentScope.end;
            scope._escapeCalled = false;
        }
        else {
            scope = new HandleScope(envObject.ctx.handleStore, currentScope.id + 1, currentScope, currentScope.end);
        }
        this.currentScope = scope;
        envObject.openHandleScopes++;
        return scope;
    }
    closeScope(envObject) {
        if (envObject.openHandleScopes === 0)
            return;
        const scope = this.currentScope;
        this.currentScope = scope.parent;
        scope.dispose();
        envObject.openHandleScopes--;
    }
    dispose() {
        let scope = this.currentScope;
        while (scope !== null) {
            scope.handleStore = null;
            scope.id = 0;
            scope.parent = null;
            scope.start = HandleStore.MIN_ID;
            scope.end = HandleStore.MIN_ID;
            scope._escapeCalled = false;
            const child = scope.child;
            scope.child = null;
            scope = child;
        }
        this.currentScope = null;
    }
}

class RefStore extends Store {
    constructor() {
        super(8);
    }
}

class DeferredStore extends Store {
    constructor() {
        super(8);
    }
}

class RefTracker {
    constructor() {
        this._next = null;
        this._prev = null;
    }
    finalize() { }
    link(list) {
        this._prev = list;
        this._next = list._next;
        if (this._next !== null) {
            this._next._prev = this;
        }
        list._next = this;
    }
    unlink() {
        if (this._prev !== null) {
            this._prev._next = this._next;
        }
        if (this._next !== null) {
            this._next._prev = this._prev;
        }
        this._prev = null;
        this._next = null;
    }
    static finalizeAll(list) {
        while (list._next !== null) {
            list._next.finalize();
        }
    }
}

class Finalizer {
    constructor(envObject, _finalizeCallback = 0, _finalizeData = 0, _finalizeHint = 0) {
        this.envObject = envObject;
        this._finalizeCallback = _finalizeCallback;
        this._finalizeData = _finalizeData;
        this._finalizeHint = _finalizeHint;
    }
    callback() { return this._finalizeCallback; }
    data() { return this._finalizeData; }
    hint() { return this._finalizeHint; }
    resetFinalizer() {
        this._finalizeCallback = 0;
        this._finalizeData = 0;
        this._finalizeHint = 0;
    }
    dispose() {
        this.envObject = undefined;
    }
}

class RefBase extends Finalizer {
    constructor(envObject, initial_refcount, ownership, finalize_callback, finalize_data, finalize_hint) {
        super(envObject, finalize_callback, finalize_data, finalize_hint);
        this._next = null;
        this._prev = null;
        this._refcount = initial_refcount;
        this._ownership = ownership;
        this.link(!finalize_callback ? envObject.reflist : envObject.finalizing_reflist);
    }
    static finalizeAll(list) {
        RefTracker.finalizeAll(list);
    }
    link(list) {
        RefTracker.prototype.link.call(this, list);
    }
    unlink() {
        RefTracker.prototype.unlink.call(this);
    }
    dispose() {
        this.unlink();
        this.envObject.dequeueFinalizer(this);
        super.dispose();
    }
    data() {
        return this._finalizeData;
    }
    ref() {
        return ++this._refcount;
    }
    unref() {
        if (this._refcount === 0) {
            return 0;
        }
        return --this._refcount;
    }
    refCount() {
        return this._refcount;
    }
    ownership() {
        return this._ownership;
    }
    finalize() {
        const ownership = this._ownership;
        const finalize_callback = this._finalizeCallback;
        const finalize_data = this._finalizeData;
        const finalize_hint = this._finalizeHint;
        this.resetFinalizer();
        this.unlink();
        let error;
        let caught = false;
        if (finalize_callback) {
            const fini = Number(finalize_callback);
            try {
                this.envObject.callFinalizer(fini, finalize_data, finalize_hint);
            }
            catch (err) {
                caught = true;
                error = err;
            }
        }
        if (ownership === 0) {
            this.dispose();
        }
        if (caught) {
            throw error;
        }
    }
}

class Env {
    constructor(ctx, makeDynCall_vppp, makeDynCall_vp) {
        this.ctx = ctx;
        this.makeDynCall_vppp = makeDynCall_vppp;
        this.makeDynCall_vp = makeDynCall_vp;
        this.openHandleScopes = 0;
        this.instanceData = null;
        this.tryCatch = new TryCatch();
        this.refs = 1;
        this.reflist = new RefTracker();
        this.finalizing_reflist = new RefTracker();
        this.destructing = false;
        this.finalizationScheduled = false;
        this.pendingFinalizers = [];
        this.lastError = {
            errorCode: 0,
            engineErrorCode: 0,
            engineReserved: 0
        };
        this._bindingMap = new WeakMap();
        this.id = 0;
    }
    static create(ctx, makeDynCall_vppp, makeDynCall_vp) {
        const env = new Env(ctx, makeDynCall_vppp, makeDynCall_vp);
        ctx.envStore.add(env);
        ctx.addCleanupHook(env, () => { env.unref(); }, 0);
        return env;
    }
    ref() {
        this.refs++;
    }
    unref() {
        this.refs--;
        if (this.refs === 0) {
            this.dispose();
        }
    }
    ensureHandle(value) {
        return this.ctx.ensureHandle(value);
    }
    ensureHandleId(value) {
        return this.ensureHandle(value).id;
    }
    clearLastError() {
        const lastError = this.lastError;
        lastError.errorCode = 0;
        lastError.engineErrorCode = 0;
        lastError.engineReserved = 0;
        return 0;
    }
    setLastError(error_code, engine_error_code = 0, engine_reserved = 0) {
        const lastError = this.lastError;
        lastError.errorCode = error_code;
        lastError.engineErrorCode = engine_error_code;
        lastError.engineReserved = engine_reserved;
        return error_code;
    }
    getReturnStatus() {
        return !this.tryCatch.hasCaught() ? 0 : this.setLastError(10);
    }
    callIntoModule(fn) {
        this.clearLastError();
        const r = fn(this);
        if (this.tryCatch.hasCaught()) {
            const err = this.tryCatch.extractException();
            throw err;
        }
        return r;
    }
    callFinalizer(cb, data, hint) {
        const f = this.makeDynCall_vppp(cb);
        const env = this.id;
        const scope = this.ctx.openScope(this);
        try {
            this.callIntoModule(() => { f(env, data, hint); });
        }
        finally {
            this.ctx.closeScope(this, scope);
        }
    }
    enqueueFinalizer(finalizer) {
        if (this.pendingFinalizers.indexOf(finalizer) === -1) {
            this.pendingFinalizers.push(finalizer);
        }
        if (!this.finalizationScheduled && !this.destructing) {
            this.finalizationScheduled = true;
            this.ref();
            _setImmediate(() => {
                this.finalizationScheduled = false;
                this.unref();
                this.drainFinalizerQueue();
            });
        }
    }
    dequeueFinalizer(finalizer) {
        const index = this.pendingFinalizers.indexOf(finalizer);
        if (index !== -1) {
            this.pendingFinalizers.splice(index, 1);
        }
    }
    drainFinalizerQueue() {
        while (this.pendingFinalizers.length > 0) {
            const refTracker = this.pendingFinalizers.shift();
            refTracker.finalize();
        }
    }
    dispose() {
        if (this.id === 0)
            return;
        this.destructing = true;
        this.drainFinalizerQueue();
        RefBase.finalizeAll(this.finalizing_reflist);
        RefBase.finalizeAll(this.reflist);
        this.tryCatch.extractException();
        this.ctx.envStore.remove(this.id);
        this.id = 0;
    }
    initObjectBinding(value) {
        const binding = {
            wrapped: 0,
            tag: null,
            data: 0
        };
        this._bindingMap.set(value, binding);
        return binding;
    }
    getObjectBinding(value) {
        if (this._bindingMap.has(value)) {
            return this._bindingMap.get(value);
        }
        return this.initObjectBinding(value);
    }
    setInstanceData(data, finalize_cb, finalize_hint) {
        if (this.instanceData) {
            this.instanceData.dispose();
        }
        this.instanceData = new RefBase(this, 0, 0, finalize_cb, data, finalize_hint);
    }
    getInstanceData() {
        return this.instanceData ? this.instanceData.data() : 0;
    }
}

class EmnapiError extends Error {
    constructor(message) {
        super(message);
        const ErrorConstructor = new.target;
        const proto = ErrorConstructor.prototype;
        if (!(this instanceof EmnapiError)) {
            const setPrototypeOf = Object.setPrototypeOf;
            if (typeof setPrototypeOf === 'function') {
                setPrototypeOf.call(Object, this, proto);
            }
            else {
                this.__proto__ = proto;
            }
            if (typeof Error.captureStackTrace === 'function') {
                Error.captureStackTrace(this, ErrorConstructor);
            }
        }
    }
}
Object.defineProperty(EmnapiError.prototype, 'name', {
    configurable: true,
    writable: true,
    value: 'EmnapiError'
});
class NotSupportWeakRefError extends EmnapiError {
    constructor(api, message) {
        super(`${api}: The current runtime does not support "FinalizationRegistry" and "WeakRef".${message ? ` ${message}` : ''}`);
    }
}
Object.defineProperty(NotSupportWeakRefError.prototype, 'name', {
    configurable: true,
    writable: true,
    value: 'NotSupportWeakRefError'
});
class NotSupportBigIntError extends EmnapiError {
    constructor(api, message) {
        super(`${api}: The current runtime does not support "BigInt".${message ? ` ${message}` : ''}`);
    }
}
Object.defineProperty(NotSupportBigIntError.prototype, 'name', {
    configurable: true,
    writable: true,
    value: 'NotSupportBigIntError'
});

class Persistent {
    constructor(value) {
        this._ref = value;
    }
    setWeak(param, callback) {
        if (this._ref === null)
            return;
        if (!supportFinalizer)
            return;
        if (this._ref instanceof WeakRef)
            return;
        this._param = param;
        this._callback = callback;
        Persistent._registry.register(this._ref, this, this);
        this._ref = new WeakRef(this._ref);
    }
    clearWeak() {
        if (this._ref === null)
            return;
        if (!supportFinalizer)
            return;
        if (this._ref instanceof WeakRef) {
            try {
                Persistent._registry.unregister(this);
            }
            catch (_) { }
            this._param = undefined;
            this._callback = undefined;
            this._ref = this._ref.deref();
        }
    }
    reset(other) {
        if (supportFinalizer) {
            try {
                Persistent._registry.unregister(this);
            }
            catch (_) { }
        }
        this._param = undefined;
        this._callback = undefined;
        if (other) {
            this._ref = other;
        }
        else {
            this._ref = null;
        }
    }
    isEmpty() {
        return this._ref === null;
    }
    deref() {
        var _a;
        if (!supportFinalizer) {
            return (_a = this._ref) !== null && _a !== void 0 ? _a : undefined;
        }
        if (this._ref === null)
            return undefined;
        if (this._ref instanceof WeakRef) {
            return this._ref.deref();
        }
        return this._ref;
    }
}
Persistent._registry = supportFinalizer
    ? new FinalizationRegistry((value) => {
        value._ref = null;
        const callback = value._callback;
        const param = value._param;
        value._callback = undefined;
        value._param = undefined;
        if (typeof callback === 'function') {
            callback(param);
        }
    })
    : undefined;

function weakCallback(ref) {
    ref.persistent.reset();
    ref.envObject.enqueueFinalizer(ref);
}
class Reference extends RefBase {
    constructor(envObject, initialRefcount, ownership, finalize_callback = 0, finalize_data = 0, finalize_hint = 0) {
        super(envObject, initialRefcount >>> 0, ownership, finalize_callback, finalize_data, finalize_hint);
        this.id = 0;
    }
    static create(envObject, handle_id, initialRefcount, ownership, finalize_callback = 0, finalize_data = 0, finalize_hint = 0) {
        const handle = envObject.ctx.handleStore.get(handle_id);
        if (!isReferenceType(handle.value)) {
            throw new TypeError('Invalid reference value');
        }
        const ref = new Reference(envObject, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint);
        envObject.ctx.refStore.add(ref);
        ref.persistent = new Persistent(handle.value);
        if (initialRefcount === 0) {
            ref._setWeak();
        }
        return ref;
    }
    ref() {
        if (this.persistent.isEmpty()) {
            return 0;
        }
        const count = super.ref();
        if (count === 1) {
            const obj = this.persistent.deref();
            if (obj) {
                this.persistent.clearWeak();
            }
        }
        return count;
    }
    unref() {
        if (this.persistent.isEmpty()) {
            return 0;
        }
        const oldRefcount = this.refCount();
        const refcount = super.unref();
        if (oldRefcount === 1 && refcount === 0) {
            const obj = this.persistent.deref();
            if (obj) {
                this._setWeak();
            }
        }
        return refcount;
    }
    get() {
        const obj = this.persistent.deref();
        if (obj) {
            const handle = this.envObject.ensureHandle(obj);
            return handle.id;
        }
        return 0;
    }
    _setWeak() {
        this.persistent.setWeak(this, weakCallback);
    }
    finalize() {
        this.persistent.reset();
        super.finalize();
    }
    dispose() {
        if (this.id === 0)
            return;
        this.persistent.reset();
        this.envObject.ctx.refStore.remove(this.id);
        super.dispose();
        this.id = 0;
    }
}

class Deferred {
    constructor(ctx, value) {
        this.id = 0;
        this.ctx = ctx;
        this.value = value;
    }
    static create(ctx, value) {
        const deferred = new Deferred(ctx, value);
        ctx.deferredStore.add(deferred);
        return deferred;
    }
    resolve(value) {
        this.value.resolve(value);
        this.dispose();
    }
    reject(reason) {
        this.value.reject(reason);
        this.dispose();
    }
    dispose() {
        this.ctx.deferredStore.remove(this.id);
        this.id = 0;
        this.value = null;
        this.ctx = null;
    }
}

class CleanupHookCallback {
    constructor(envObject, fn, arg, order) {
        this.envObject = envObject;
        this.fn = fn;
        this.arg = arg;
        this.order = order;
    }
}
class CleanupQueue {
    constructor() {
        this._cleanupHooks = [];
        this._cleanupHookCounter = 0;
    }
    empty() {
        return this._cleanupHooks.length === 0;
    }
    add(envObject, fn, arg) {
        if (this._cleanupHooks.filter((hook) => (hook.envObject === envObject && hook.fn === fn && hook.arg === arg)).length > 0) {
            throw new Error('Can not add same fn and arg twice');
        }
        this._cleanupHooks.push(new CleanupHookCallback(envObject, fn, arg, this._cleanupHookCounter++));
    }
    remove(envObject, fn, arg) {
        for (let i = 0; i < this._cleanupHooks.length; ++i) {
            const hook = this._cleanupHooks[i];
            if (hook.envObject === envObject && hook.fn === fn && hook.arg === arg) {
                this._cleanupHooks.splice(i, 1);
                return;
            }
        }
    }
    drain() {
        const hooks = this._cleanupHooks.slice();
        hooks.sort((a, b) => (b.order - a.order));
        for (let i = 0; i < hooks.length; ++i) {
            const cb = hooks[i];
            if (typeof cb.fn === 'number') {
                cb.envObject.makeDynCall_vp(cb.fn)(cb.arg);
            }
            else {
                cb.fn(cb.arg);
            }
            this._cleanupHooks.splice(this._cleanupHooks.indexOf(cb), 1);
        }
    }
    dispose() {
        this._cleanupHooks.length = 0;
        this._cleanupHookCounter = 0;
    }
}
class NodejsWaitingRequestCounter {
    constructor() {
        this.refHandle = new MessageChannel().port1;
        this.count = 0;
    }
    increase() {
        if (this.count === 0) {
            this.refHandle.ref();
        }
        this.count++;
    }
    decrease() {
        if (this.count === 0)
            return;
        if (this.count === 1) {
            this.refHandle.unref();
        }
        this.count--;
    }
}
class Context {
    constructor() {
        this.envStore = new EnvStore();
        this.scopeStore = new ScopeStore();
        this.refStore = new RefStore();
        this.deferredStore = new DeferredStore();
        this.handleStore = new HandleStore();
        this.cbinfoStack = new CallbackInfoStack();
        this.feature = {
            supportReflect,
            supportFinalizer,
            supportBigInt,
            supportNewFunction,
            canSetFunctionName,
            setImmediate: _setImmediate,
            Buffer
        };
        this.cleanupQueue = new CleanupQueue();
        if (typeof process === 'object' && process !== null && typeof process.once === 'function') {
            this.refCounter = new NodejsWaitingRequestCounter();
            process.once('beforeExit', () => {
                this.runCleanup();
            });
        }
    }
    createNotSupportWeakRefError(api, message) {
        return new NotSupportWeakRefError(api, message);
    }
    createNotSupportBigIntError(api, message) {
        return new NotSupportBigIntError(api, message);
    }
    createReference(envObject, handle_id, initialRefcount, ownership, finalize_callback = 0, finalize_data = 0, finalize_hint = 0) {
        return Reference.create(envObject, handle_id, initialRefcount, ownership, finalize_callback, finalize_data, finalize_hint);
    }
    createDeferred(value) {
        return Deferred.create(this, value);
    }
    createEnv(makeDynCall_vppp, makeDynCall_vp) {
        return Env.create(this, makeDynCall_vppp, makeDynCall_vp);
    }
    getCurrentScope() {
        return this.scopeStore.currentScope;
    }
    addToCurrentScope(value) {
        return this.scopeStore.currentScope.add(value);
    }
    openScope(envObject) {
        return this.scopeStore.openScope(envObject);
    }
    closeScope(envObject, _scope) {
        this.scopeStore.closeScope(envObject);
    }
    ensureHandle(value) {
        switch (value) {
            case undefined: return HandleStore.UNDEFINED;
            case null: return HandleStore.NULL;
            case true: return HandleStore.TRUE;
            case false: return HandleStore.FALSE;
            case _global: return HandleStore.GLOBAL;
        }
        return this.addToCurrentScope(value);
    }
    addCleanupHook(envObject, fn, arg) {
        this.cleanupQueue.add(envObject, fn, arg);
    }
    removeCleanupHook(envObject, fn, arg) {
        this.cleanupQueue.remove(envObject, fn, arg);
    }
    runCleanup() {
        while (!this.cleanupQueue.empty()) {
            this.cleanupQueue.drain();
        }
    }
    increaseWaitingRequestCounter() {
        var _a;
        (_a = this.refCounter) === null || _a === void 0 ? void 0 : _a.increase();
    }
    decreaseWaitingRequestCounter() {
        var _a;
        (_a = this.refCounter) === null || _a === void 0 ? void 0 : _a.decrease();
    }
    dispose() {
        this.runCleanup();
    }
}
let defaultContext;
function createContext() {
    return new Context();
}
function getDefaultContext() {
    if (!defaultContext) {
        defaultContext = createContext();
    }
    return defaultContext;
}

const version = "0.34.0";

export { CallbackInfo, CallbackInfoStack, ConstHandle, Context, Deferred, DeferredStore, EmnapiError, Env, EnvStore, Finalizer, Handle, HandleScope, HandleStore, NotSupportBigIntError, NotSupportWeakRefError, Persistent, RefBase, RefStore, RefTracker, Reference, ScopeStore, Store, TryCatch, createContext, getDefaultContext, isReferenceType, version };
